#include <xen/errno.h>
#include <xen/lib.h>
#include <xen/smp.h>
#include <xen/sched.h>
#include <xen/dmi.h>
#include <asm/uaccess.h>
#include <asm/io.h>

#ifdef __XEN_COMPAT_H
#undef __XEN_COMPAT_H
#endif
#include <public/platform.h>

extern void host_request_smi(unsigned int dest);

static long cc_cr4(xenpf_stratus_call_t *cc_call) {
	int rw = cc_call->u.cr4.rw;
	unsigned long cr4;

	if (rw) {	// Write
		return -ENOSYS;
	} else {	// Read
		asm("movq %%cr4,%0" : "=r" (cr4));
		cc_call->u.cr4.cr4 = cr4;
	}

	return 0;
}

static long cc_cpuid(xenpf_stratus_call_t *cc_call) {
	cpuid(  cc_call->u.cpuid.op,
		&cc_call->u.cpuid.eax,
		&cc_call->u.cpuid.ebx,
		&cc_call->u.cpuid.ecx,
		&cc_call->u.cpuid.edx );

	return 0;
}

static long cc_rw_msr(xenpf_stratus_call_t *cc_call) {
	if (cc_call->u.msr.rw == 0) {
		// Read
		rdmsrl(cc_call->u.msr.msr, cc_call->u.msr.val);
	} else {
		wrmsrl(cc_call->u.msr.msr, cc_call->u.msr.val);
	}

	return 0;
}

static long cc_lapic_id(xenpf_stratus_call_t *cc_call) {
	cc_call->u.ls.id = GET_APIC_ID(apic_read(APIC_ID));
	return 0;
}

#define DUMP_VECTOR_PHYS  	(0xf00)
#define HOST_BIOS_VECTOR_PHYS	(0xff0)
#define HOST_BIOS_VECTOR_SIZE	(0x10)


static long cc_rw_hbv(xenpf_stratus_call_t *cc_call) {
	int rw = cc_call->u.rw.rw;
	int size = cc_call->u.rw.size;
	unsigned long where = cc_call->u.rw.where;

	if (((unsigned long)where + (unsigned long)size) > 
			HOST_BIOS_VECTOR_SIZE) {
		return -EFAULT;
	}

	where += (unsigned long)__va(HOST_BIOS_VECTOR_PHYS);

	if (rw) {	// Write
		if (copy_from_user((void*)where, cc_call->u.rw.data, size)) {
			return -EFAULT;
		}
	} else {	// Read
		if (copy_to_user(cc_call->u.rw.data, (void*)where, size)) {
			return -EFAULT;
		}
	}

	return 0;
}

static long cc_rw_dumpvec(xenpf_stratus_call_t *cc_call) {
	int rw = cc_call->u.rw.rw;
	int size = cc_call->u.rw.size;
	void *where = __va(DUMP_VECTOR_PHYS);

	if (size > sizeof(int)) {
		return -EINVAL;
	}

	if (rw) {	// Write
		if (copy_from_user((void*)where, cc_call->u.rw.data, size)) {
			return -EFAULT;
		}
	} else {	// Read
		if (copy_to_user(cc_call->u.rw.data, (void*)where, size)) {
			return -EFAULT;
		}
	}

	return 0;
}

static long cc_rw_region(xenpf_stratus_call_t *cc_call) {
	switch (cc_call->u.rw.region) {
		case RW_HBV:
			return cc_rw_hbv(cc_call);
			break;
		case RW_DUMPVEC:
			return cc_rw_dumpvec(cc_call);
			break;
		default:
			return -EINVAL;
	}
}

static long cc_smi(xenpf_stratus_call_t *cc_call) {
	host_request_smi(cc_call->u.smi.dest);
	return 0;
}

static long cc_hbv_memset(xenpf_stratus_call_t *cc_call) {
	int size = cc_call->u.hbv_m.size;

	if (size > HOST_BIOS_VECTOR_SIZE)
		size = HOST_BIOS_VECTOR_SIZE;

	memset(__va(HOST_BIOS_VECTOR_PHYS), cc_call->u.hbv_m.val, cc_call->u.hbv_m.size);

	return 0;
}

static int locked_out = 1;

static int found_stratus(struct dmi_system_id *d)
{
	printk("Stratus platform detected.\n");
	return 0;
}

#define NO_MATCH	{ DMI_NONE, NULL}
#define MATCH		DMI_MATCH

static struct dmi_system_id stratus_platform[] = {
	{ found_stratus, "Stratus Platform", {
			MATCH(DMI_BOARD_VENDOR, "Stratus"),
			NO_MATCH, NO_MATCH, NO_MATCH
			} },
	{ NULL, NULL, }
};

static int check_stratus_dmi(void) {
	// Run dmi scan looking for Stratus Vendor string.
	if (dmi_check_system(stratus_platform))
		return 0;

	return 1;
}

long do_stratus(xenpf_stratus_call_t *call) {
	long ret = -EINVAL;

	if (!IS_PRIV(current->domain))
		return -EPERM;

	if (call->cmd == CC_VALIDATE_PLATFORM)
		locked_out = check_stratus_dmi();

	if (locked_out)
		return -EPERM;

	switch (call->cmd) {
	case CC_TRIGGER_SMI:
		ret = cc_smi(call);
		break;
	case CC_HBV_MEMSET:
		ret = cc_hbv_memset(call);
		break;
	case CC_RW_REGION:
		ret = cc_rw_region(call);
		break;
	case CC_LAPIC_ID:
		ret = cc_lapic_id(call);
		break;
	case CC_CR4:
		ret = cc_cr4(call);
		break;
	case CC_CPUID:
		ret = cc_cpuid(call);
		break;
	case CC_RW_MSR:
		ret = cc_rw_msr(call);
		break;
	case CC_VALIDATE_PLATFORM:
		ret = 0;	// If we made it here, we are on a Stratus box.
		break;
	default:
		printk("%s:line %d, unknown command %d\n", __func__,
			__LINE__, call->cmd);
		break;
	}

	call->ret = ret;

	return ret;
}

